#
# Copyright (c) 2022 QASS GmbH.
# Website: https://qass.net
# Contact: QASS GmbH <info@qass.net>
#
# This file is part of Qass tools
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from typing import List, Sequence, Tuple, Union
import re
import warnings
from sqlalchemy import (
    Float,
    create_engine,
    Column,
    Integer,
    String,
    BigInteger,
    Identity,
    Index,
    Enum,
    TypeDecorator,
    select,
    text,
)
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql.selectable import Select
from pathlib import Path
from enum import Enum
from tqdm.auto import tqdm
from multiprocessing.pool import Pool as Pool

from .buffer_parser import Buffer, InvalidFileError


__all__ = ["BufferMetadataCache", "BufferMetadata"]


class BufferEnum(TypeDecorator):
    impl = String

    def __init__(self, enumtype, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumtype = enumtype

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.name

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self._enumtype[str(value)]


__Base = declarative_base()


class BufferMetadata(__Base):
    """This class acts as a template for buffer files. It's properties represent all available metadata of a buffer file.
    This class is used internally as a database model and can be instantiated to provide a template for a buffer file by
    populating desired properties and passing the object to the cache which will in turn create a query based on this object.
    """

    __tablename__ = "buffer_metadata"
    properties = (
        "id",
        "project_id",
        "directory_path",
        "filename",
        "header_size",
        "process",
        "channel",
        "datamode",
        "datakind",
        "datatype",
        "process_time",
        "process_date_time",
        "db_header_size",
        "bytes_per_sample",
        "db_count",
        "full_blocks",
        "db_size",
        "db_sample_count",
        "frq_bands",
        "db_spec_count",
        "compression_frq",
        "compression_time",
        "avg_time",
        "avg_frq",
        "spec_duration",
        "frq_start",
        "frq_end",
        "frq_per_band",
        "sample_count",
        "spec_count",
        "adc_type",
        "bit_resolution",
        "fft_log_shift",
        "streamno",
        "preamp_gain",
        "analyzer_version",
        "partnumber",
        "header_hash",
    )

    id = Column(Integer, Identity(start=1), primary_key=True)
    project_id = Column(BigInteger, index=True)
    directory_path = Column(String, nullable=False, index=True)
    filename = Column(String, nullable=False)
    header_hash = Column(String(64), index=True)
    machine_id = Column(String, nullable=True)
    header_size = Column(Integer)
    process = Column(Integer, index=True)
    channel = Column(Integer, index=True)
    datamode = Column(
        BufferEnum(Buffer.DATAMODE), index=True
    )  # this is an ENUM in buffer_parser
    datakind = Column(BufferEnum(Buffer.DATAKIND))  # this is an ENUM in buffer_parser
    datatype = Column(BufferEnum(Buffer.DATATYPE))  # this is an ENUM in buffer_parser
    process_time = Column(BigInteger)
    process_date_time = Column(String)
    db_header_size = Column(Integer)
    bytes_per_sample = Column(Integer)
    db_count = Column(Integer)
    full_blocks = Column(Integer)
    db_size = Column(Integer)
    db_sample_count = Column(Integer)
    frq_bands = Column(Integer)
    db_spec_count = Column(Integer)
    compression_frq = Column(Integer, index=True)
    compression_time = Column(Integer, index=True)
    avg_time = Column(Integer, index=True)
    avg_frq = Column(Integer, index=True)
    spec_duration = Column(Float)
    frq_start = Column(Integer)
    frq_end = Column(Integer)
    frq_per_band = Column(Float)
    sample_count = Column(BigInteger)
    spec_count = Column(BigInteger)
    adc_type = Column(
        BufferEnum(Buffer.ADCTYPE)
    )  # TODO this is an ENUM in buffer_parser
    bit_resolution = Column(Integer)
    fft_log_shift = Column(Integer)
    streamno = Column(Integer)
    preamp_gain = Column(Integer)
    analyzer_version = Column(String)
    partnumber = Column(String)

    opening_error = Column(String, nullable=True)

    @hybrid_property
    def filepath(self):
        return str(Path(str(self.directory_path)) / self.filename)

    @staticmethod
    def buffer_to_metadata(buffer):
        """Converts a Buffer object to a BufferMetadata database object by copying all the @properties from the Buffer
        object putting them in the BufferMetadata object

        :param buffer: Buffer object
        :type buffer: buffer_parser.Buffer
        """
        file = Path(buffer.filepath)
        directory_path = str(file.parent)
        buffer_metadata = BufferMetadata(
            filename=file.name, directory_path=directory_path
        )
        for prop in BufferMetadata.properties:
            try:  # try to map all the buffer properties and skip on error
                setattr(
                    buffer_metadata, prop, getattr(buffer, prop)
                )  # get the @property method and execute it
            except Exception:
                continue
        return buffer_metadata


Index("project_id_process_index", BufferMetadata.project_id, BufferMetadata.process)
Index(
    "project_id_process_channel_index",
    BufferMetadata.project_id,
    BufferMetadata.process,
    BufferMetadata.channel,
)
Index(
    "compression_time_frq_index",
    BufferMetadata.compression_time,
    BufferMetadata.compression_frq,
)
Index(
    "project_id_compression_time_frq_index",
    BufferMetadata.project_id,
    BufferMetadata.compression_time,
    BufferMetadata.compression_frq,
)


def _create_metadata(args):
    """Create a metadata object from a Buffer object by trying to open
    the buffer file contained in the args tuple"""
    Buffer_cls, file = args
    try:
        with Buffer_cls(file.resolve()) as buffer:
            buffer_metadata = BufferMetadata.buffer_to_metadata(buffer)
            return buffer_metadata
    except InvalidFileError:
        pass
    except Exception as e:
        directory_path, filename = str(file.parent), file.name
        warnings.warn(f"One or more Buffers couldn't be opened {file}", UserWarning)
        return BufferMetadata(
            directory_path=directory_path, filename=filename, opening_error=str(e)
        )


class BufferMetadataCache:
    """This class acts as a Cache for Buffer Metadata. It uses a database session with a buffer_metadata table to map
    metadata to files on the disk. The cache can be queried a lot faster than manually opening a lot of buffer files.
    """

    BufferMetadata = BufferMetadata

    def __init__(self, db_url="sqlite:///:memory:", Buffer_cls=Buffer):
        self.engine = create_engine(db_url)
        BufferMetadata.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)
        self.Buffer_cls = Buffer_cls

    def synchronize_directory(
        self,
        *paths,
        sync_subdirectories=True,
        regex_pattern="^.*[p][0-9]*[c][0-9]{1}[b][0-9]{2}",
        verbose=1,
        delete_stale_entries=False,
        machine_id=None,
        glob_pattern="*p*c?b*",
    ):
        """synchronize the buffer files in the given paths with the database matching the regex pattern

        :param paths: The absolute paths to the directory
        :type paths: str
        :param recursive: When True synchronize all of the subdirectories recursively, defaults to True
        :type recursive: bool, optional
        :param regex_pattern: The regex pattern validating the buffer naming format (matched on file.name)
        :type regex_pattern: string, optional
        :param verbose: verbosity level. 0 = no feedback, 1 = progress bar
        :type verbose: int, optional
        :param machine_id: An optional identifier for a certain machine to enable synchronization of different platforms
        :type machine_id: string, optional
        :param glob_pattern: The pattern forwarded to Path.glob. This pattern acts as a preselection for the
            files retrieved for the regex pattern
        :type glob_pattern: string, optional
        """
        pattern = re.compile(regex_pattern)
        for path in paths:
            if sync_subdirectories:
                files = [
                    file
                    for file in Path(path).resolve().rglob(glob_pattern)
                    if file.is_file() and pattern.match(file.name)
                ]
            else:
                files = [
                    file
                    for file in Path(path).resolve().glob(glob_pattern)
                    if file.is_file() and pattern.match(file.name)
                ]
            unsynchronized_files, synchronized_missing_buffers = (
                self.get_non_synchronized_files(files, machine_id)
            )
            if delete_stale_entries:
                self.remove_files_from_cache(
                    synchronized_missing_buffers, verbose=verbose
                )
            self.add_files_to_cache(
                unsynchronized_files,
                verbose=verbose,
                machine_id=machine_id,
                check_synced=False,
            )

    def get_matching_metadata(self, query: Select) -> List[BufferMetadata]:
        """Query the cache for all BufferMetadata database entries matching

        :param query: A sqlalchemy select statement specifying the properties of the BufferMetadata objects
        :type query: Select
        :return: A list with the matching BufferMetadata objects
        :rtype: list[BufferMetadata]
        """
        with self.Session() as session:
            matching_metadata = session.scalars(query).all()
        return matching_metadata

    def get_matching_files(self, query: Select) -> List[str]:
        """Query the Cache for all files matching the properties that selected by the query object.
        The usage of the buffer_metadata, filter_functions and sort_key is deprecated and will be removed in
        two minor versions. Use the sqlalchemy query parameter instead.

        .. code-block:: python
                :linenos:

                BufferMetadataCache.get_matching_files(
                    select(BM).filter(BM.channel==1, BM.compression_freq==4, BM.process > 100)
                )
                # Returns all buffer filepaths with channel = 1, A frequency compression of 4,
                # processes above 100 sorted by the process number

        :param query: A sqlalchemy select statement specifying the properties of the BufferMetadata objects
        :type query: Select
        :return: A list with the paths to the buffer files that match the buffer_metadata
        :rtype: list[str]
        """
        matching_metadata = self.get_matching_metadata(query)
        return [str(m.filepath) for m in matching_metadata]

    def get_matching_buffers(self, query: Select) -> List[Buffer]:
        """Calls get_matching_files and converts the result to Buffer objects

        :return: List of Buffer objects
        :rtype: list
        """
        files = self.get_matching_files(query)
        buffers = []
        for file in files:
            try:
                with self.Buffer_cls(file) as buffer:
                    buffers.append(buffer)
            except Exception:
                pass
        return buffers

    def get_non_synchronized_files(
        self,
        files: Sequence[Path],
        machine_id: Union[str, None] = None,
    ) -> Tuple[List[Path], List[Path]]:
        """calculate the difference between the set of files and the set of synchronized files

        :param files: filenames
        :type files: Sequence[Path]
        :param machine_id: machine identifier
        :type machine_id: Union[str, None]
        :return: The set of files that are not synchronized, and the database entries that exist but the file is not present anymore
        """
        file_set = set(files)
        with self.Session() as session:
            synchronized_buffers = set(
                Path(str(buffer.filepath))
                for buffer in session.query(self.BufferMetadata)
                .filter(BufferMetadata.machine_id == machine_id)
                .all()
            )
        unsynchronized_files = [
            Path(p) for p in file_set.difference(synchronized_buffers)
        ]
        synchronized_missing_buffers = [
            Path(p) for p in synchronized_buffers.difference(file_set)
        ]
        return unsynchronized_files, synchronized_missing_buffers

    def add_files_to_cache(
        self,
        files: Sequence[Path],
        verbose: int = 0,
        batch_size: int = 1000,
        machine_id: Union[str, None] = None,
        check_synced: bool = True,
    ):
        """Add buffer files to the cache by providing the complete filepaths
        If a file (determined by filename and directory) is already synchronized it will be skipped

        :param files: complete filepaths that are added to the cache. The filepath is used with the Buffer class to open a buffer and extract the header information.
        :type files: list, tuple of str
        :param verbose: verbosity level. 0 = no feedback, 1 = progress bar
        :type verbose: int, optional
        :param batch_size: The batch size after which the cache will commit a batch to the database
        :type batch_size: int, optional
        :param machine_id: A unique identifier for a different machine
        :type machine_id: str, optional
        :param check_synced: Whether to check if the files that are about to be added are already synced to the cache
        :type check_synced: bool, optional
        """
        if check_synced:
            files, _ = self.get_non_synchronized_files(files, machine_id)
        with self.Session() as session:
            with Pool() as pool:
                params = [(self.Buffer_cls, f) for f in files]
                f_gen = pool.imap_unordered(_create_metadata, params)
                f_gen = (
                    tqdm(f_gen, desc="Adding Buffers", total=len(params))
                    if verbose > 0 and len(params) > 0
                    else f_gen
                )
                for i, metadata in enumerate(filter(lambda m: m is not None, f_gen)):
                    metadata: BufferMetadata
                    assert metadata is not None, "BufferMetadata object is None"
                    metadata.machine_id = machine_id
                    session.add(metadata)
                    if i % batch_size == 0:
                        session.commit()
            session.commit()

    def remove_files_from_cache(self, files: List[Path], verbose=0):
        """Remove synchronized files from the cache

        :param files: complete filepaths that are present in the cache
        :type files: list, tuple of Path
        :param verbose: verbosity level. 0 = no feedback, 1 = progress bar
        :type verbose: int, optional
        """
        with self.Session() as session:
            file_iter = (
                tqdm(files, desc="Removing File Entries")
                if verbose > 0 and len(files) > 0
                else files
            )
            for file in file_iter:
                try:
                    with self.Buffer_cls(file) as b:
                        pass
                    entry = (
                        session.query(self.BufferMetadata)
                        .filter(self.BufferMetadata.header_hash == b.header_hash)
                        .one()
                    )
                    if not entry:
                        continue
                    session.delete(entry)
                except Exception as e:
                    session.rollback()
                    raise e
            session.commit()

    def get_buffer_metadata_query(self, buffer_metadata):
        """Converts a .. py:class:: BufferMetadata object to a complete query. Every property of the object will be converted into
        SQL and returned as a ..py:class:: sqlalchemy.orm.query.FromStatement object

        :param buffer_metadata: The template BufferMetadata object.
        :type buffer_metadata: BufferMetadata
        :return: The sqlalchemy query object
        :rtype: sqlalchemy.orm.query.FromStatement
        """
        q = "SELECT * FROM buffer_metadata WHERE opening_error IS NULL AND "
        for prop in self.BufferMetadata.properties:
            prop_value = getattr(buffer_metadata, prop)
            if prop_value is not None:
                if isinstance(prop_value, str):
                    prop_value = f"'{prop_value}'"
                elif isinstance(prop_value, Enum):
                    prop_value = f"'{prop_value.name}'"
                q += f"{prop} = {prop_value} AND "
        q = q[:-4]  # prune the last AND
        return select(BufferMetadata).from_statement(text(q))


def get_declarative_base():
    """Getter for the declarative Base that is used by the :py:class:`BufferMetadataCache`.

    :return: declarative base class
    """
    return __Base



#
# Copyright (c) 2022 QASS GmbH.
# Website: https://qass.net
# Contact: QASS GmbH <info@qass.net>
#
# This file is part of Qass tools
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import json
from uuid import uuid4

from qass.tools.analyzer.buffer_metadata_cache import (
    BufferMetadataCache as BMC,
    BufferMetadata as BM,
)
from qass.tools.analyzer.buffer_parser import Buffer
from qass.tools.analyzer.testing import MockBuffer
from sqlalchemy import inspect, func, select
import pytest

SEED = 42


@pytest.fixture(scope="function")
def cache():
    """Returns a BufferMetadataCache instance with an in memory database"""
    cache = BMC(Buffer_cls=MockBuffer)  # type: ignore
    return cache


def test_session_creation():
    cache = BMC()
    inspector = inspect(cache.engine)
    assert "buffer_metadata" in inspector.get_table_names()


@pytest.mark.parametrize(
    "filepath,data",
    [
        (
            "foo/bar/hoo.000",
            {
                "directory_path": "foo/bar",
                "filename": "hoo.000",
                "project_id": 0,
                "header_hash": uuid4(),
                "header_size": 2000,
                "process": 0,
                "channel": 0,
                "datamode": Buffer.DATAMODE.DATAMODE_FFT,
            },
        ),
        (
            "foo/bar.000",
            {
                "directory_path": "foo",
                "filename": "bar.000",
                "project_id": 2,
                "header_hash": uuid4(),
                "header_size": 4000,
                "process": 1234,
                "channel": 1,
                "datamode": Buffer.DATAMODE.DATAMODE_FFT,
                "bit_resolution": 2,
                "compression_time": 4,
                "compression_frq": 4,
            },
        ),
    ],
)
def test_buffer_to_metadata(filepath, data):
    mock_buffer = MockBuffer(filepath, **data)
    metadata = BM.buffer_to_metadata(mock_buffer)
    assert metadata.directory_path == mock_buffer.directory_path  # type: ignore
    assert metadata.filename == mock_buffer.filename  # type: ignore
    assert metadata.header_hash == mock_buffer.header_hash  # type: ignore
    assert metadata.process == mock_buffer.process  # type: ignore
    assert metadata.channel == mock_buffer.channel  # type: ignore
    assert metadata.datamode == mock_buffer.datamode  # type: ignore


@pytest.mark.parametrize(
    "filenames,datas",
    [
        (["foo"], [{"process": 1}]),
        (["foo", "bar", "hoo"], [{}, {}, {}]),
        (["foo", "bar", "foo"], [{}, {}, {}]),
        (
            ["foo", "bar", "hoo"],
            [
                {"process": 2, "compression_time": 1},
                {"header_hash": str(uuid4())},
                {"compression_time": 4, "compression_frq": 8},
            ],
        ),
    ],
)
def test_add_files_to_cache(tmp_path, cache, filenames, datas):
    files = [tmp_path / filename for filename in filenames]
    unique_files = set(files)
    for file, data in zip(files, datas):
        if file.exists():
            # skip duplicate files
            continue
        with open(file, "w") as f:
            json.dump(data, f)
    cache.add_files_to_cache(files)
    with cache.Session() as session:
        for filename, data in zip(filenames, datas):
            bm = session.query(BM).filter(BM.filename == filename).first()
            assert bm is not None, f"File {filename} is missing from the cache"
            for key, value in data.items():
                assert getattr(bm, key) == value, (
                    f"key: {key} did not match: {getattr(bm, key), value}"
                )
        assert session.query(func.count(BM.id)).scalar() == len(unique_files)


def test_add_files_to_cache_no_check_synced(tmp_path, cache):
    file = tmp_path / "a"
    with open(file, "w") as f:
        json.dump({"header_hash": file.name}, f)
    cache.add_files_to_cache([file] * 10, check_synced=False)
    with cache.Session() as session:
        assert session.query(func.count(BM.id)).scalar() == 10


def test_add_files_to_cache_invalid_file(tmp_path, cache):
    # Invalid files should be ignored silently
    file = tmp_path / "invalid.txt"
    with open(file, "w") as f:
        f.write("invalid data")
    cache.add_files_to_cache([file])
    with cache.Session() as session:
        bm = session.query(BM).first()
        assert bm is None


@pytest.mark.parametrize(
    "synced, new_hashes, machine_id, r_unsynced, r_stale",
    [
        ([BM(filename="0", header_hash="0")], ["0"], None, [], []),
        (
            [BM(filename="0", header_hash="0"), BM(filename="1", header_hash="1")],
            ["0"],
            None,
            [],
            ["1"],
        ),
        (
            [BM(filename="0", header_hash="0"), BM(filename="1", header_hash="1")],
            ["0", "2", "3"],
            None,
            ["2", "3"],
            ["1"],
        ),
        (
            [],
            ["0", "2", "3"],
            None,
            ["0", "2", "3"],
            [],
        ),
        (
            [
                BM(filename="0", header_hash="0", machine_id="0.0.0.0"),
                BM(filename="1", header_hash="1"),
            ],
            ["0", "1", "2", "3"],
            "0.0.0.0",
            ["1", "2", "3"],
            [],
        ),
        (
            [BM(filename=str(i), header_hash=str(i)) for i in range(200)],
            [str(i) for i in range(100)],
            None,
            [],
            [str(i) for i in range(100, 200)],
        ),
    ],
)
def test_get_non_synchronized_files(
    tmp_path, cache, synced, new_hashes, machine_id, r_unsynced, r_stale
):
    with cache.Session() as session:
        for bm in synced:
            bm.directory_path = str(tmp_path.resolve())
            session.add(bm)
        session.commit()

    new_files = []
    for hash_ in new_hashes:
        # name files same as their hashes
        test_file = tmp_path / hash_
        with open(test_file, "w") as f:
            json.dump({"header_hash": hash_}, f)
        new_files.append(test_file)
    unsynced, stale = cache.get_non_synchronized_files(new_files, machine_id=machine_id)
    unsynced_files = [tmp_path / hash_ for hash_ in r_unsynced]
    stale_files = [tmp_path / hash_ for hash_ in r_stale]
    assert len(set(unsynced).difference(set(unsynced_files))) == 0
    assert len(set(stale).difference(set(stale_files))) == 0


def test_get_non_synchronized_files_invalid_files(tmp_path, cache):
    file = tmp_path / "invalid.txt"
    with open(file, "w") as f:
        f.write("invalid data")
    _, _ = cache.get_non_synchronized_files([file])
    assert True


def test_synchronize_directory(tmp_path, cache):
    N = 100
    files = [tmp_path / str(i) for i in range(N)]
    for file in files:
        with open(file, "w") as f:
            json.dump({"header_hash": file.stem}, f)
    cache.synchronize_directory(tmp_path, glob_pattern="*", regex_pattern=".*")
    with cache.Session() as session:
        synced_files = session.query(func.count(BM.id)).scalar()
    assert synced_files == N


def test_remove_files_from_cache(tmp_path, cache):
    synced = [
        BM(filename="0", directory_path="./", header_hash="0"),
        BM(filename="1", directory_path="./", header_hash="1"),
        BM(filename="2", directory_path="./", header_hash="2"),
    ]
    with cache.Session() as session:
        for bm in synced:
            session.add(bm)
        session.commit()
        assert session.query(func.count(BM.id)).scalar() == 3
    test_file = tmp_path / "1"
    with open(test_file, "w") as f:
        json.dump({"header_hash": "1"}, f)
    cache.remove_files_from_cache([test_file])
    with cache.Session() as session:
        assert session.query(func.count(BM.id)).scalar() == 2
        synced_bm = session.query(BM).all()
        for bm in synced_bm:
            assert bm.header_hash in ("0", "2")


def test_get_matching_metadata(cache):
    synced = [
        BM(filename="0", directory_path="./", header_hash="0", process=1),
        BM(filename="1", directory_path="./", header_hash="1", process=2),
        BM(filename="2", directory_path="./", header_hash="2", process=3),
    ]
    with cache.Session() as session:
        for bm in synced:
            session.add(bm)
        session.commit()
        assert session.query(func.count(BM.id)).scalar() == 3
        results = cache.get_matching_metadata(select(BM).filter(BM.process < 3))
        assert len(results) == 2


def test_get_matching_buffers(tmp_path, cache):
    data = [
        {"header_hash": "0", "process": 0},
        {"header_hash": "1", "process": 1},
        {"header_hash": "2", "process": 2},
        {"header_hash": "3", "process": 3},
    ]
    for d in data:
        file = tmp_path / d["header_hash"]
        with file.open(mode="w") as f:
            json.dump(d, f)
    cache.synchronize_directory(tmp_path, glob_pattern="*", regex_pattern=".*")
    buffers = cache.get_matching_buffers(select(BM))
    assert len(buffers) == 4


def test_get_buffer_metadata_query(cache):
    synced = [
        BM(filename="0", directory_path="./", header_hash="0", process=1),
        BM(filename="1", directory_path="./", header_hash="1", process=2),
        BM(filename="2", directory_path="./", header_hash="2", process=3),
    ]
    with cache.Session() as session:
        for bm in synced:
            session.add(bm)
        session.commit()
    match_result = cache.get_matching_metadata(select(BM).filter(BM.process == 2))
    query_result = cache.get_matching_metadata(
        cache.get_buffer_metadata_query(BM(process=2))
    )
    match_hashes = set(map(lambda bm: bm.header_hash, match_result))
    query_hashes = set(map(lambda bm: bm.header_hash, query_result))
    assert len(match_hashes.difference(query_hashes)) == 0



#
# Copyright (c) 2022 QASS GmbH.
# Website: https://qass.net
# Contact: QASS GmbH <info@qass.net>
#
# This file is part of Qass tools 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import glob
from  qass.tools.analyzer.deleting_process import DeleteHandler
from pathlib import Path
import os
import shutil
import pytest
from unittest.mock import Mock

# mocked version
@pytest.fixture
def deletehandler_obj_helper(tmp_path):
    pattern = "*.txt"
    obj = DeleteHandler(tmp_path, pattern)
    return obj

@pytest.fixture
def deletehandler_obj_helper_two(tmp_path):
    pattern = "*.txt"
    obj = DeleteHandler(tmp_path, pattern, log_entries=True)
    return obj

@pytest.fixture
def get_oldest_helper(monkeypatch):
    def get_oldest_wrapper(self, some_list):
        test_list = [('Tue Jul  5 11:47:07 2022', '/home/opti/tmp_dir/Text File.txt'), ('Tue Jul  5 11:47:14 2022', '/home/opti/tmp_dir/Text File (1).txt'), ('Tue Jun 14 16:49:34 2022', '/home/opti/tmp_dir/file1.txt')]
        return test_list
    
    monkeypatch.setattr(DeleteHandler, "_DeleteHandler__get_oldest", get_oldest_wrapper)

def test_delete_by_amount(monkeypatch, deletehandler_obj_helper, get_oldest_helper):
    # Arrange
    amount_limit = 1
    os_mock = Mock()
    len_mock = Mock(return_value=[0,1,2])
    monkeypatch.setattr(glob, "glob",len_mock)
    monkeypatch.setattr(os, "remove", os_mock)
    
    # Act
    deletehandler_obj_helper.delete_by_amount(amount_limit)
    
    # Assert
    assert os_mock.call_count == 2

def test_delete_by_disk_space(monkeypatch, deletehandler_obj_helper, get_oldest_helper):
    # Arrange
    usage_limit = 0.9

    os_mock = Mock()
    shutil_mock = Mock(return_value=(1000, 950, 50))
    filesize_mock = Mock(return_value=25)
    len_mock = Mock(return_value=[0,1,2])
    monkeypatch.setattr(shutil, "disk_usage", shutil_mock)
    monkeypatch.setattr(os, "remove", os_mock)
    monkeypatch.setattr(glob, "glob",len_mock)
    monkeypatch.setattr(os.path, "getsize", filesize_mock)
    
    # Act
    deletehandler_obj_helper.delete_by_disk_space(usage_limit)
    
    # Assert
    assert os_mock.call_count == 2

def test_delete_by_amount_two(monkeypatch, deletehandler_obj_helper_two, tmp_path, get_oldest_helper):
    # Arrange
    amount_limit = 1
    os_mock = Mock()
    len_mock = Mock(return_value=[0,1,2])
    monkeypatch.setattr(glob, "glob",len_mock)
    monkeypatch.setattr(os, "remove", os_mock)
    expected_path = tmp_path / Path("*.txt" + ".log")
    
    # Act
    deletehandler_obj_helper_two.delete_by_amount(amount_limit)
    
    # Assert
    assert os_mock.call_count == 2
    assert os.path.exists(expected_path)

def test_delete_by_disk_space_two(monkeypatch, deletehandler_obj_helper_two, tmp_path, get_oldest_helper):
    # Arrange
    usage_limit = 0.9

    os_mock = Mock()
    shutil_mock = Mock(return_value=(1000, 950, 50))
    filesize_mock = Mock(return_value=25)
    len_mock = Mock(return_value=[0,1,2])
    monkeypatch.setattr(shutil, "disk_usage", shutil_mock)
    monkeypatch.setattr(os, "remove", os_mock)
    monkeypatch.setattr(glob, "glob",len_mock)
    monkeypatch.setattr(os.path, "getsize", filesize_mock)
    expected_path = tmp_path / Path("*.txt" + ".log")
    # Act
    deletehandler_obj_helper_two.delete_by_disk_space(usage_limit)
    
    # Assert
    assert os_mock.call_count == 2
    assert os.path.exists(expected_path)


#
# Copyright (c) 2022 QASS GmbH.
# Website: https://qass.net
# Contact: QASS GmbH <info@qass.net>
#
# This file is part of Qass tools
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Virtual devices

This module provides a simplified interface for the virtual devices of the Analyzer4D software.
The Analyzer4D software provides an interface that allows to add customized devices that are not handled by the measurement hardware.
But the data coming from those virtual devices will be handled in the same way like the time domain signal from the measurement hardware.
The data is stored in a so called "Buffer" that works based on equally distanced data.
The resulting data streams are completely compatible to other data streams in the software.

Data channels from a virtual device have to be configured in the Analyzer4D software in the multiplexer settings (Configuration->Multiplexer->Virtual input configuration).
There the devices can be assigned to measurement ports.
"""

from Analyzer.Devices import VirtDeviceInterface
from Analyzer.Core import Log_IF

# to not block the Analyzer4D software when reading data each virtual device works in its own thread.
from PySide2.QtCore import QThread, QByteArray, Slot, QElapsedTimer
from PySide2.QtWidgets import QDialog
from threading import Lock

from abc import ABC, abstractmethod
import traceback
from typing import List, Dict, Tuple
import json

from sys import version_info
if version_info.major != 3:
    raise RuntimeError('Expected Python3')
if version_info.minor <= 6:
    from collections import Sequence
else:
    from collections.abc import Sequence


class VirtualInputDevice(ABC):
    """
    VirtualInputDevice is a simplified interface for custom virtual devices in the Analyzer4D software.
    The methods marked with "abstractmethod" must be implemented.
    The other methods can be overridden if needed.

    This class is meant to be used by DeviceHandler (see below).
    The device handler implements the actual Analyzer4D interface class while using instances of this class for the communication.
    """

    def __init__(self,
                 name: str,
                 sample_rate: float,
                 normal_amp: float,
                 request_rate: float=100,
                 dont_close: bool=False):
        """instantiate a VirtualInputDevice.
        This class is an abstract base class.
        You must derive from this class and implement the abstract funtions to create a custom device.

        :param name: This is the name of the device that will also be displayed in the GUI.
        The name is used as an identifier and thus it must be unique inside one DeviceTypeCollection.
        :type name: str

        :param sample_rate: This is the sample rate that is used by the Analyzer4D software to build the Data signal stream.
        All values are assumed to have the same distance in time.
        :type sample_rate: float

        :param normal_amp: This is the maximum value that this device will provide.
        This is used by the Analyzer4D software for the correct scaling.
        :type normal_amp: float

        :param request_rate: How often the get_data() should be called?, defaults to 100 - meaning 100 calls of get_data() in a second.
        If zero or None the device communication thread will not sleep but only yieldCurrentThread.
        This might cause issues since the python interpreter is heavily used by this interface.
        :type request_rate: float, optional

        :param dont_close: If True the function open_connection() will be called as early as possible and close_connection() will be called as late as possible.
        The device connection will not be closed between different processes.
        If False the method open_connection() will be called at the beginning of each measurement and close_connection() will be called at the end of each measurement.
        Defaults to False.
        :type dont_close: bool, optional
        """

        self._config = {
            'sample_rate': sample_rate,
            'normal_amp': normal_amp,
            'request_rate': request_rate,
            'dont_close': dont_close
            }

        self._name = name

    @abstractmethod
    def is_available(self) -> bool:
        """is_available should indicate whether its possible to establish a connection to the device.

        :return: True if the device is available and ready for measurements, False otherwise.
        :rtype: bool
        """
        ...

    @abstractmethod
    def open_connection(self):
        """open_connection() will be called before a measurement.
        The communication to the device should be setup in this function.
        It is assumed that its save to call get_data after this function returned without error.
        This function is called at each begin of a measurement.
        It is called even when the flag "dont_stop" is True.
        Thus you can reestablish a damaged connection if needed.

        raise an exception if needed. It will be handled by the wrapping class DeviceHandler.
        """
        ...

    @abstractmethod
    def close_connection(self):
        """close_connection() will be called to close the connection to the device (usually at the end of a measurement).
        The communication to the device should be completely cleaned up in this function.
        It is assumed that its save to call open_connection() again after this function returned without error.

        raise an exception if needed. It will be handled by the wrapping class DeviceHandler.
        """
        ...

    @abstractmethod
    def get_data(self) -> List[float]:
        """Fetch the data from the device and convert it to a List of floats.
        Note: Ensure that the data type is float and not np.float or anything else!
        This method should always clear the device's buffer containing the data when it's called.
        This is to prevent old data from persisting.

        :return: The list of new values. The list must not contain old values!
        :rtype: List[float]
        """
        ...

    # overide the functions below if needed
    def init_device(self):
        return None

    def config_dialog(self):
        """If this function returns a QDialog this will be displayed in the Analyzer4D software to setup the device.
        The settings button will only be available in the software if at least one device provides a QDialog and not None in this funciton.

        :return: your custom config dialog. Defaults to None.
        :rtype: _type_
        """
        return None

    @property
    def requests_per_sec(self):
        return self._config['request_rate']

    @property
    def sample_rate(self):
        return self._config['sample_rate']

    @property
    def normal_amplitude(self):
        return self._config['normal_amp']

    @property
    def dont_close(self):
        return self._config['dont_close']

    @property
    def name(self):
        return self._name

    @property
    def stream_count(self):
        return 1 if 'stream_configs' not in self._config else len(self._config['stream_configs'])

    @property
    def stream_configs(self):
        return None if 'stream_configs' not in self._config else self._config['stream_configs']

    def get_config(self):
        return self._config

    def set_config(self, config):
        required_keys = ("request_rate", "sample_rate", "normal_amp", "dont_close")
        for key in required_keys:
            if key not in config.keys():
                raise KeyError(f"Config key missing: {key}")
        self._config = config

    def apply_config(self):
        pass


class MultiStreamVirtualInputDevice(VirtualInputDevice):
    def __init__(self,
                 name: str,
                 stream_configs: Tuple[Dict[str, int]],
                 request_rate: float=100,
                 dont_close: bool=False):
        """instantiate a MultiStreamVirtualInputDevice.
        This class is an abstract base class and inherits from VirtualInputDevice.
        You must derive from this class and implement the abstract funtions to create a custom device.

        :param name: This is the name of the device that will also be displayed in the GUI.
        The name is used as an identifier and thus it must be unique inside one DeviceTypeCollection.
        :type name: str

        :param stream_configs: This is a list of configs for each stream.
        The list's length defines the number of streams provided by this device.
        The config must include at least the following keywords:
        "stream_name": A string to identify the stream (displayed)
        "sample_rate": float: The streams sample rate in Hz
        "normal_amp": The maximum expected value (used for display scaling)
        :type stream_configs: Tuple[Dict[str, int]]

        :param request_rate: How often the get_data() should be called?, defaults to 100 - meaning 100 calls of get_data() in a second.
        If zero or None the device communication thread will not sleep but only yieldCurrentThread.
        This might cause issues since the python interpreter is heavily used by this interface.
        :type request_rate: float, optional

        :param dont_close: If True the function open_connection() will be called as early as possible and close_connection() will be called as late as possible.
        The device connection will not be closed between different processes.
        If False the method open_connection() will be called at the beginning of each measurement and close_connection() will be called at the end of each measurement.
        Defaults to False.
        :type dont_close: bool, optional
        """

        # The sample rate must not be 0 even if it is not used!
        super().__init__(name, 1, 0, request_rate, dont_close)

        required_keys = ("stream_name", "sample_rate", "normal_amp")
        filter_req = lambda x: x not in required_keys

        for sc in stream_configs:
            filtered_keys = list(filter(filter_req, sc))
            if filtered_keys:
                raise ValueError(f'At least one entry of stream_configs does not contain required keys {",".join(filtered_keys)}')

        self._config['stream_configs'] = stream_configs

    @abstractmethod
    def get_data(self) -> Tuple[List[float]]:
        """Fetch the data for all streams and convert them to a Tuple of List of floats.
        The tuple is expected to have stream_count elements.
        Note: Ensure that the data type is float and not np.float or anything else!
        This method should always clear the device's buffer containing the data when it's called.
        This is to prevent old data from persisting.

        :return: The lists of new values for all streams, packed into one tuple.
        The list must not contain old values!
        :rtype: Tuple[List[float]]
        """
        ...


class _DeviceThread(QThread):
    """_DeviceThread implements a QThread.
    The thread is startet when a measurement starts.
    Inside this thread communication to the device takes place.
    """
    def __init__(self, device: VirtualInputDevice):
        super().__init__()

        self.should_stop = False
        self.values = []
        self.lock = Lock()
        self.device = device

    def fetch_values(self, stream: int) -> List[float]:
        """fetch_values is called by the interface from the reading thread in the Analyzer4D software.
        It first aqcuires the lock to prevent race conditions when fetching the values.

        :return: The list of currently available values.
        :rtype: List[float]
        """
        with self.lock:
            vals = self.values[stream].copy()
            self.values[stream].clear()

        return vals

    def stop(self):
        self.should_stop = True

    def run(self):
        """The run method is implicitly called by the QThread class.
        It is executed in a seperate thread.
        The thread finishes when this function terminates.

        Exceptions caught during the execution will be logged as a popupError.
        This should lead to a fail state in the Analyzer4D software, including clearing the 'ready' output line.
        """
        try:
            request_rate = self.device.requests_per_sec

            if request_rate:
                elapsed = QElapsedTimer()
                elapsed.start()
                read_counter = 0

            self.should_stop = False
            with self.lock:
                self.values = tuple(list() for _ in range(self.device.stream_count))

            self.device.open_connection()

            # empty the data queue at start - we do not want to use old data
            if self.device.dont_close:
                self.device.get_data()


            while not self.should_stop:
                new_data = self.device.get_data()
                if not isinstance(new_data, Sequence):
                    raise ValueError(f'Device {self.device.name} get_data must return a Sequence but returned {type(new_data)}')

                if new_data:
                    first_elem = new_data[0]
                    if not isinstance(first_elem, Sequence):
                        if self.device.stream_count != 1:
                            raise ValueError(f'Device {self.device.name} get_data must return a Sequence of Sequences since it has multiple streams.')
                        with self.lock:
                            self.values[0].extend(new_data)
                    else:
                        with self.lock:
                            for idx, d in enumerate(new_data):
                                self.values[idx].extend(d)

                if request_rate:
                    read_counter += 1
                    elapsed_should = 1/request_rate * 1e9 * read_counter
                    elapsed_is = elapsed.nsecsElapsed()
                    wait_time = int((elapsed_should - elapsed_is) / 1e3)
                    if wait_time > 0:
                        QThread.usleep(wait_time)
                else:
                    self.yieldCurrentThread()

            if not self.device.dont_close:
                self.device.close_connection()
        except Exception:
            Log_IF.popupError(f'Exception caught in device thread:\n{traceback.format_exc()}')


class DeviceTypeCollection(VirtDeviceInterface):
    """DeviceClass implements the interface VirtDeviceInterface from the Analyzer4D software.
    The functions of this interface are called by the Analyzer4D software.
    This class is a plugin implementation of the interface.

    A DeviceClass is a class of hardware devices.
    Usually all devices handled by this class work similar to each other and support the same features.
    """
    def __init__(self, devices: List[VirtualInputDevice], name:str='', version:str=''):
        """
        The constructor expects a .
        The devices are expected to be of the type VirtualInputDevice.

        :param devices: List of the device. Their names are used as identifiers and have to be unique inside this DeviceTypeCollection.
        :type devices: List[VirtualInputDevice]
        :param name: The name of this device class, defaults to ''
        :type name: str, optional
        :param version: The version string of this device class implementation, defaults to ''
        :type version: str, optional
        """
        super().__init__()

        self._name = name
        self._version = version

        self._devices = {}
        for dev in devices:
            if dev.name in self._devices:
                raise ValueError('The devices` names have to be unique, but got multiple devices with the same name')
            self._devices[dev.name] = dev

        self._device_threads: Dict[_DeviceThread] = {}
        for name, dev in self._devices.items():
            self._device_threads[name] = _DeviceThread(dev)

    def init(self):
        """
        This function is called by the Analyzer4D software.
        Usually device detection takes part here.
        We assume in this implementatoin that available devices have been passed in the constructor.
        """
        pass

    def capabilities(self) -> Dict:
        """capabilites is called by the Analyzer4D software to get information about available devices and their configuration.
        Especially the sample rate and the capability of configuration dialogs is queried here.
        The config dict consists of two main keys (inputs, configdialog)
        The 'inputs' key contains an interface config with the keys (ifaceName : str, sampleRate: float, normalAmplitude: float)
        The 'configdialog' is a boolean indicating whether a QDialog will be displayed
        Missing keys in the config will

        :return: A dictionary containing information about this device class and the devices themselves.
        :rtype: Dict
        """
        interfaces_conf = []

        has_config = False
        for name, dev in self._devices.items():
            if dev.config_dialog() is not None:
                has_config = True

            int_conf = {
                'ifaceName': name,
                'sampleRate': dev.sample_rate,
                'normalAmplitude': dev.normal_amplitude,
                'streamcount': dev.stream_count
            }

            if dev.stream_configs:
                for idx, sc in enumerate(dev.stream_configs):
                    int_conf[f'streamname{idx}'] = sc["stream_name"]
                    int_conf[f'streamrate{idx}'] = sc["sample_rate"]
                    int_conf[f'streamampl{idx}'] = sc["normal_amp"]

            interfaces_conf.append(int_conf)

        config = {
            'inputs': interfaces_conf,
            'configdialog': has_config
        }
        return config

    def versionString(self) -> str:
        """versionString

        :return: The version of this device class implementation.
        :rtype: str
        """
        return self._version

    def name(self) -> str:
        """name

        :return: This device class' name.
        :rtype: str
        """
        return self._name

    def isInputAvailable(self, name: str) -> bool:
        """isInputAvailable requests to check whether the given device is available and ready for measurements.

        :param name: The device's identifier string
        :type name: str
        :return: True if the device is available and ready. False otherwise.
        :rtype: bool
        """
        if name not in self._devices:
            return False

        return self._devices[name].is_available()

    def inputCount(self) -> int:
        stream_count = sum(dev.stream_count for dev in self._devices)
        return stream_count

    def openInput(self, name: str) -> bool:
        """openInput starts the communication and the data fetching for the given device.
        This function starts the device's thread.
        All communication to the device and the data collection is done inside this thread.

        :param name: The device's identifier string
        :type name: str
        :return: True on success, False otherwise
        :rtype: bool
        """
        try:
            if not self._device_threads[name].isRunning():
                self._device_threads[name].start()
            return True
        except Exception as e:
            Log_IF.popupError(f'Exception caught during opening virtual device {name}:\n{traceback.format_exc()}')
            return False

    def closeInput(self, name: str, prepareRestart = False) -> bool:
        """closeInput requests to stop the data collection for the given device.
        In this implementation this causes the device's thread to terminate.

        :param name: The device's identifier string
        :type name: str
        :param prepareRestart: defaults to False
        :type prepareRestart: bool, optional
        :return: True on success, False otherwise
        :rtype: bool
        """
        if prepareRestart and self._device_threads[name].isRunning():
            return True

        try:
            self._device_threads[name].stop()
            deadline = 100 # ms
            if not self._device_threads[name].wait(deadline):
                Log_IF.popupError(f'Failed to close virtual device {name} within {deadline} ms - Timeout.')
                return False

            return True
        except Exception as e:
            Log_IF.popupError(f'Exception caught during closing virtual device {name}:\n{traceback.format_exc()}')
            return False

    def readInput(self, name: str, stream: int) -> List[float]:
        """readInput fetches and returns the currently available data from the given device's thread.

        :param name: The device's identifier string
        :type name: str
        :param stream:
        :type stream: int, optional
        :return: List of available new values.
        :rtype: List[float]
        """
        vals = self._device_threads[name].fetch_values(stream)
        return vals

    def getConfig(self, name: str) -> QByteArray:
        """The Analyzer4D software calls this function to make settings persistent in the database.
        This implementation converts a dictionary based configuration from the device to the QByteArray that is expected by the Analyzer4D software.

        :param name: The device's identifier string
        :type name: str
        :return: The byte array - the configuration's textual representation
        :rtype: QByteArray
        """
        try:
            config_dict = self._devices[name].get_config()
            json_str = json.dumps(config_dict)
            return json_str.encode()
        except Exception as e:
            Log_IF.popupError(f'Exception caught during getting config from virtual device {name}:\n{traceback.format_exc()}')

    def setConfig(self, name: str, data: QByteArray) -> bool:
        """setConfig is called after the device initialization to propagate saved settings from the database to the devices themselves.
        This function translates the settings from a QByteArray representation to a json dictionary.

        :param name: The device's identifier string
        :type name: str
        :param data: The byte array containing the textual representation of the device.
        :type data: QByteArray
        :return: True on success, False otherwise (if an exception has been raised)
        :rtype: bool
        """
        try:
            # There is a very nasty bug in older PySide2 / Python versions - The following crashes!
            # We have to do the conversion differently.
            #json_str = bytes(data).decode()

            json_str = ''
            for c in data:
                json_str += c.decode()

            if json_str:
                config_dict = json.loads(json_str)
                self._devices[name].set_config(config_dict)
            return True
        except Exception as e:
            Log_IF.popupError(f'Exception caught during setting config to virtual device {name}:\n{traceback.format_exc()}')
            return False

    def applyConfig(self, name: str) -> bool:
        """applyConfig is called by the Analyzer4D software to request the device to apply the config.
        This might include to communicate settings to a device.

        :param name: The device's identifier string
        :type name: str
        :return: True on success, False otherwise (if an exception has been raised)
        :rtype: bool
        """
        try:
            self._devices[name].apply_config()
            return True
        except Exception as e:
            Log_IF.popupError(f'Exception caught during config application of virtual device {name}:\n{traceback.format_exc()}')
            return False

    def createPluginDialog(self) -> QDialog:
        """createPluginDialog allows to implement a custom dialog for this device type.
        This implementation does not use this feature.
        But you can override this function by subclassing this class...

        :return: A QDialog instance.
        :rtype: QDialog
        """
        return None

    def createConfigDialog(self, name: str) -> QDialog:
        """createConfigDialog is called by the Analyzer4D software to allow the user to configurate the device.
        This mitght include file paths, IP addresses, Ports, sample rates or other device specific parameters.
        If the device implementation supports this feature it must return a QDialog instance in its implementation of config_dialog().
        This dialog is just forwarded here.

        :param name: The device's identifier string
        :type name: str
        :return: The configuration QDialog instance.
        :rtype: QDialog
        """
        try:
            return self._devices[name].config_dialog()
        except Exception as e:
            Log_IF.popupError(f'Exception caught during creation of config dialog for virtual device {name}:\n{traceback.format_exc()}')
            return None



#
# Copyright (c) 2022 QASS GmbH.
# Website: https://qass.net
# Contact: QASS GmbH <info@qass.net>
#
# This file is part of Qass tools
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#


#
# Copyright (c) 2022 QASS GmbH.
# Website: https://qass.net
# Contact: QASS GmbH <info@qass.net>
#
# This file is part of Qass tools 
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
from typing import Callable
import warnings
from PySide2.QtCore import QObject, Signal, Slot, Qt, SIGNAL, QThread
from PySide2.QtGui import QGuiApplication

try:
    from Analyzer.OutputRedirect import OutputRedirect_Manager
    import Analyzer.Debugger
except:
    warnings.warn("Analyzer modules couldn't be imported, are you inside the Analyzer environment?", RuntimeWarning)


__all__ = ["inMainThread"]

class inMainThread(QObject):
    def __init__(self, synchron=True):
        super().__init__()
        self.synchron = synchron
        # move self QObject to the main thread
        main_thread = QGuiApplication.instance().thread()
        self.moveToThread(main_thread)

    called = Signal()
    
    def __call__(self, method: Callable):
        """
        Invokes a method on the main thread. Taking care of garbage collection "bugs".
        """
        def inner_func(*args, **kwargs):
            if QThread.currentThread() == QGuiApplication.instance().thread():
                return method(*args, **kwargs)
            
            self._method = method
            
            # do not garbage collect this self object until the function has finished.
            self.setParent(QGuiApplication.instance())
            
            self.params = args
            self.kwargs = kwargs
            
            self.stdout_redirector = OutputRedirect_Manager.getInstance().getCurrentOutRedirect()
            self.stderr_redirector = OutputRedirect_Manager.getInstance().getCurrentErrRedirect()
            self.debuggers = Analyzer.Debugger.DebuggerManager.getInstance().getCurrDebuggers()
            
            if self.synchron:
                # This function must not be called from the main thread - this would result in a deadlock
                QObject.disconnect(self, SIGNAL('called()'), self.execute)
                self.called.connect(self.execute, Qt.BlockingQueuedConnection)
                self.called.emit()
                # the result will only be available in the synchronous execution mode

                if self._exc_info is not None:
                    type, exc, traceback = self._exc_info
                    raise exc
                return self._result
            else:
                QObject.disconnect(self, SIGNAL('called()'), self.execute)
                self.called.connect(self.execute)
                self.called.emit()
        
        return inner_func

    @Slot()
    def execute(self):
        OutputRedirect_Manager.getInstance().registerStdOutHandler(self.stdout_redirector)
        OutputRedirect_Manager.getInstance().registerStdErrHandler(self.stderr_redirector)
        
        for dbg in self.debuggers:
            Analyzer.Debugger.DebuggerManager.getInstance().registerDebugger(dbg)
        
        self._result = None
        self._exc_info = None
        
        try:
            # actually execute the wrapped method
            self._result = self._method(*self.params, **self.kwargs)
        except Exception as e:
            import sys
            self._exception = e
            self._exc_info = sys.exc_info()
        finally:
            for dbg in self.debuggers:
                Analyzer.Debugger.DebuggerManager.getInstance().unregisterDebugger(dbg)

            OutputRedirect_Manager.getInstance().unregisterStdOutHandler(self.stdout_redirector)
            OutputRedirect_Manager.getInstance().unregisterStdErrHandler(self.stderr_redirector)
            # decref self -> trigger garbage collector
            self.setParent(None)
